import merge from 'deepmerge';
import { startService as esbuildStartService } from 'esbuild';
import { EventEmitter } from 'events';
import { promises as fs } from 'fs';
import glob from 'glob';
import * as colors from 'kleur/colors';
import mkdirp from 'mkdirp';
import path from 'path';
import rimraf from 'rimraf';
import { generateEnvModule, wrapHtmlResponse, wrapImportMeta, wrapImportProxy, } from '../build/build-import-proxy';
import { buildFile } from '../build/build-pipeline';
import { createImportResolver } from '../build/import-resolver';
import srcFileExtensionMapping from '../build/src-file-extension-mapping';
import { removeLeadingSlash, removeTrailingSlash } from '../config';
import { stopEsbuild } from '../plugins/plugin-esbuild';
import { transformFileImports } from '../rewrite-imports';
import { printStats } from '../stats-formatter';
import { getEncodingType, getExt, replaceExt } from '../util';
import { getInstallTargets, run as installRunner } from './install';
import { paint } from './paint';
async function installOptimizedDependencies(allFilesToResolveImports, installDest, commandOptions) {
    console.log(colors.yellow('! optimizing dependencies...'));
    const installConfig = merge(commandOptions.config, {
        installOptions: {
            dest: installDest,
            env: { NODE_ENV: process.env.NODE_ENV || 'production' },
            treeshake: commandOptions.config.installOptions.treeshake ?? true,
        },
    });
    // 1. Scan imports from your final built JS files.
    const installTargets = await getInstallTargets(installConfig, Object.values(allFilesToResolveImports));
    // 2. Install dependencies, based on the scan of your final build.
    const installResult = await installRunner({
        ...commandOptions,
        installTargets,
        config: installConfig,
    });
    // 3. Print stats immediate after install output.
    if (installResult.stats) {
        console.log(printStats(installResult.stats));
    }
    return installResult;
}
export async function command(commandOptions) {
    const { cwd, config } = commandOptions;
    const messageBus = new EventEmitter();
    const isBundledHardcoded = config.devOptions.bundle !== undefined;
    const bundlerPlugin = config._bundler;
    const isBundled = isBundledHardcoded ? !!config.devOptions.bundle : !!bundlerPlugin;
    const bundlerDashboardId = bundlerPlugin ? bundlerPlugin.name : 'bundle';
    const buildDirectoryLoc = isBundled ? path.join(cwd, `.build`) : config.devOptions.out;
    const internalFilesBuildLoc = path.join(buildDirectoryLoc, config.buildOptions.metaDir);
    const finalDirectoryLoc = config.devOptions.out;
    if (config.buildOptions.clean)
        rimraf.sync(buildDirectoryLoc);
    mkdirp.sync(buildDirectoryLoc);
    mkdirp.sync(internalFilesBuildLoc);
    if (finalDirectoryLoc !== buildDirectoryLoc) {
        if (config.buildOptions.clean)
            rimraf.sync(finalDirectoryLoc);
        mkdirp.sync(finalDirectoryLoc);
    }
    console.log = (...args) => {
        messageBus.emit('CONSOLE', { level: 'log', args });
    };
    console.warn = (...args) => {
        messageBus.emit('CONSOLE', { level: 'warn', args });
    };
    console.error = (...args) => {
        messageBus.emit('CONSOLE', { level: 'error', args });
    };
    let relDest = path.relative(cwd, config.devOptions.out);
    if (!relDest.startsWith(`..${path.sep}`)) {
        relDest = `.${path.sep}` + relDest;
    }
    paint(messageBus, config.plugins.map((p) => p.name), { dest: relDest }, undefined);
    if (!isBundled) {
        messageBus.emit('WORKER_UPDATE', {
            id: bundlerDashboardId,
            state: ['SKIP', 'dim'],
        });
    }
    for (const runPlugin of config.plugins) {
        if (runPlugin.run) {
            messageBus.emit('WORKER_START', { id: runPlugin.name });
            runPlugin
                .run({
                isDev: false,
                log: (msg, data) => {
                    messageBus.emit(msg, { ...data, id: runPlugin.name });
                },
            })
                .then(() => {
                messageBus.emit('WORKER_COMPLETE', { id: runPlugin.name, error: null });
            })
                .catch((err) => {
                messageBus.emit('WORKER_COMPLETE', { id: runPlugin.name, error: err });
            });
        }
    }
    let minifierService = await esbuildStartService(); // create esbuild process for minifying (there may or may not be one in plugins)
    // Write the `import.meta.env` contents file to disk
    await fs.writeFile(path.join(internalFilesBuildLoc, 'env.js'), generateEnvModule('production'));
    const includeFileSets = [];
    for (const [fromDisk, toUrl] of Object.entries(config.mount)) {
        const dirDisk = path.resolve(cwd, fromDisk);
        const dirDest = path.resolve(buildDirectoryLoc, removeLeadingSlash(toUrl));
        const allFiles = glob.sync(`**/*`, {
            ignore: config.exclude,
            cwd: dirDisk,
            absolute: true,
            nodir: true,
            dot: true,
        });
        const allBuildNeededFiles = [];
        await Promise.all(allFiles.map(async (f) => {
            f = path.resolve(f); // this is necessary since glob.sync() returns paths with / on windows.  path.resolve() will switch them to the native path separator.
            allBuildNeededFiles.push(f);
        }));
        includeFileSets.push([dirDisk, dirDest, allBuildNeededFiles]);
    }
    const allBuiltFromFiles = new Set();
    const allFilesToResolveImports = {};
    for (const [dirDisk, dirDest, allFiles] of includeFileSets) {
        for (const locOnDisk of allFiles) {
            const { baseExt: fileExt } = getExt(locOnDisk);
            let outLoc = locOnDisk.replace(dirDisk, dirDest);
            let builtLocOnDisk = locOnDisk;
            const extToReplace = config._extensionMap[fileExt] || srcFileExtensionMapping[fileExt];
            if (extToReplace) {
                outLoc = replaceExt(outLoc, extToReplace);
                builtLocOnDisk = replaceExt(builtLocOnDisk, extToReplace);
            }
            const builtFileOutput = await buildFile(locOnDisk, {
                buildPipeline: config.plugins,
                messageBus,
                isDev: false,
            });
            allBuiltFromFiles.add(locOnDisk);
            const { baseExt, expandedExt } = getExt(outLoc);
            let contents = builtFileOutput[config._extensionMap[fileExt] || srcFileExtensionMapping[fileExt] || fileExt];
            if (!contents) {
                continue;
            }
            const cssOutPath = outLoc.replace(/.js$/, '.css');
            mkdirp.sync(path.dirname(outLoc));
            switch (baseExt) {
                case '.js': {
                    if (builtFileOutput['.css']) {
                        await fs.mkdir(path.dirname(cssOutPath), { recursive: true });
                        await fs.writeFile(cssOutPath, builtFileOutput['.css'], 'utf-8');
                        contents = `import './${path.basename(cssOutPath)}';\n` + contents;
                    }
                    contents = wrapImportMeta({ code: contents, env: true, isDev: false, hmr: false, config });
                    // minify install if enabled and there’s no bundler
                    if (config.buildOptions.minify && !config._bundler) {
                        const { js } = await minifierService.transform(contents, { minify: true });
                        if (js)
                            contents = js;
                    }
                    allFilesToResolveImports[outLoc] = { baseExt, expandedExt, contents, locOnDisk };
                    break;
                }
                case '.html': {
                    contents = wrapHtmlResponse({
                        code: contents,
                        isDev: false,
                        hmr: false,
                        config,
                    });
                    allFilesToResolveImports[outLoc] = { baseExt, expandedExt, contents, locOnDisk };
                    break;
                }
            }
            await fs.writeFile(outLoc, contents, getEncodingType(baseExt));
        }
    }
    stopEsbuild();
    minifierService.stop();
    const installDest = path.join(buildDirectoryLoc, config.buildOptions.webModulesUrl);
    const installResult = await installOptimizedDependencies(allFilesToResolveImports, installDest, commandOptions);
    if (!installResult.success || installResult.hasError) {
        process.exit(1);
    }
    const allImportProxyFiles = new Set();
    for (const [outLoc, file] of Object.entries(allFilesToResolveImports)) {
        const resolveImportSpecifier = createImportResolver({
            fileLoc: file.locOnDisk,
            dependencyImportMap: installResult.importMap,
            isBundled,
            config,
        });
        const resolvedCode = await transformFileImports(file, (spec) => {
            // Try to resolve the specifier to a known URL in the project
            let resolvedImportUrl = resolveImportSpecifier(spec);
            if (!resolvedImportUrl) {
                return spec;
            }
            // We treat ".proxy.js" files special: we need to make sure that they exist on disk
            // in the final build, so we mark them to be written to disk at the next step.
            const isAbsoluteUrlPath = resolvedImportUrl.startsWith('/');
            const isProxyImport = resolvedImportUrl.endsWith('.proxy.js');
            if (isProxyImport) {
                if (isAbsoluteUrlPath) {
                    allImportProxyFiles.add(path.resolve(cwd, config.devOptions.out, removeLeadingSlash(resolvedImportUrl)));
                }
                else {
                    allImportProxyFiles.add(path.resolve(path.dirname(outLoc), resolvedImportUrl));
                }
            }
            // When dealing with an absolute import path, we need to honor the baseUrl
            if (isAbsoluteUrlPath) {
                return removeTrailingSlash(config.buildOptions.baseUrl) + resolvedImportUrl;
            }
            return resolvedImportUrl;
        });
        await fs.mkdir(path.dirname(outLoc), { recursive: true });
        await fs.writeFile(outLoc, resolvedCode);
    }
    for (const importProxyFileLoc of allImportProxyFiles) {
        const originalFileLoc = importProxyFileLoc.replace('.proxy.js', '');
        const proxiedExt = path.extname(originalFileLoc);
        const proxiedCode = await fs.readFile(originalFileLoc, getEncodingType(proxiedExt));
        const proxiedUrl = originalFileLoc.substr(buildDirectoryLoc.length).replace(/\\/g, '/');
        const proxyCode = await wrapImportProxy({
            url: proxiedUrl,
            code: proxiedCode,
            isDev: false,
            hmr: false,
            config,
            messageBus,
        });
        await fs.writeFile(importProxyFileLoc, proxyCode, getEncodingType('.js'));
    }
    if (!isBundled) {
        messageBus.emit('WORKER_COMPLETE', { id: bundlerDashboardId, error: null });
        messageBus.emit('WORKER_UPDATE', {
            id: bundlerDashboardId,
            state: ['SKIP', isBundledHardcoded ? 'dim' : 'yellow'],
        });
        if (!isBundledHardcoded) {
            messageBus.emit('WORKER_MSG', {
                id: bundlerDashboardId,
                level: 'log',
                msg: `"plugins": ["@snowpack/plugin-webpack"]\n\n` +
                    `Connect a bundler plugin to optimize your build for production.\n` +
                    colors.dim(`Set "devOptions.bundle" configuration to false to remove this message.`),
            });
        }
    }
    else if (!bundlerPlugin) {
        messageBus.emit('WORKER_COMPLETE', {
            id: bundlerDashboardId,
            error: new Error('No bundler plugin connected.'),
        });
    }
    else {
        try {
            messageBus.emit('WORKER_UPDATE', { id: bundlerDashboardId, state: ['RUNNING', 'yellow'] });
            await bundlerPlugin.bundle({
                srcDirectory: buildDirectoryLoc,
                destDirectory: finalDirectoryLoc,
                jsFilePaths: allBuiltFromFiles,
                log: (msg) => {
                    messageBus.emit('WORKER_MSG', { id: bundlerDashboardId, level: 'log', msg });
                },
            });
            messageBus.emit('WORKER_COMPLETE', { id: bundlerDashboardId, error: null });
        }
        catch (err) {
            messageBus.emit('WORKER_MSG', {
                id: bundlerDashboardId,
                level: 'error',
                msg: err.toString(),
            });
            messageBus.emit('WORKER_COMPLETE', { id: bundlerDashboardId, error: err });
        }
    }
    if (finalDirectoryLoc !== buildDirectoryLoc) {
        rimraf.sync(buildDirectoryLoc);
    }
}
