import type HttpProxy from 'http-proxy';
import { ValidatorResult } from 'jsonschema';
import { Plugin as RollupPlugin } from 'rollup';
export declare type EnvVarReplacements = Record<string, string | number | true>;
export declare type SnowpackBuildMap = Record<string, string>;
/** Standard file interface */
export interface SnowpackSourceFile {
    /** base extension (e.g. `.js`) */
    baseExt: string;
    /** file contents */
    contents: string;
    /** expanded extension (e.g. `.proxy.js` or `.module.css`) */
    expandedExt: string;
    /** if no location on disk, assume this exists in memory */
    locOnDisk: string;
}
export interface LoadOptions {
    filePath: string;
    fileExt: string;
    isDev: boolean;
    log: (msg: any, data: any) => void;
}
export interface TransformOptions {
    filePath: string;
    fileExt: string;
    contents: string;
    isDev: boolean;
    log: (msg: any, data: any) => void;
}
export interface PluginProxyOptions {
    fileUrl: string;
    contents: string;
    isDev: boolean;
    log: (msg: any, data: any) => void;
}
export interface RunOptions {
    isDev: boolean;
    log: (msg: any, data: any) => void;
}
/** DEPRECATED */
export declare type __OldBuildResult = {
    result: string;
    resources?: {
        css?: string;
    };
};
/** map of extensions -> code (e.g. { ".js": "[code]", ".css": "[code]" }) */
export declare type LoadResult = string | {
    [fileExtension: string]: string;
};
export interface BundleOptions {
    srcDirectory: string;
    destDirectory: string;
    jsFilePaths: Set<string>;
    log: (msg: any, level?: 'INFO' | 'WARN' | 'ERROR') => void;
}
export interface SnowpackPlugin {
    /** name of the plugin */
    name: string;
    resolve?: {
        /** file extensions that this load function takes as input (e.g. [".jsx", ".js", …]) */
        input: string[];
        /** file extensions that this load function outputs (e.g. [".js", ".css"]) */
        output: string[];
    };
    /** load a file that matches resolve.input */
    load?(options: LoadOptions): Promise<LoadResult | null | undefined | void>;
    /** transform a file that matches resolve.input */
    transform?(options: TransformOptions): Promise<string | null | undefined | void>;
    /** controls how a non-JS file should be imported into JS. */
    proxy?(options: PluginProxyOptions): string | null | undefined | void;
    /** runs a command, unrelated to file building (e.g. TypeScript, ESLint) */
    run?(options: RunOptions): Promise<unknown>;
    /** bundle the entire built application */
    bundle?(options: BundleOptions): Promise<void>;
    /** Known dependencies that should be installed */
    knownEntrypoints?: string[];
}
export declare type ProxyOptions = HttpProxy.ServerOptions & {
    on: Record<string, Function>;
};
export declare type Proxy = [string, ProxyOptions];
export interface SnowpackConfig {
    install: string[];
    extends?: string;
    exclude: string[];
    knownEntrypoints: string[];
    webDependencies?: {
        [packageName: string]: string;
    };
    proxy: Proxy[];
    mount: Record<string, string>;
    alias: Record<string, string>;
    scripts: Record<string, string>;
    plugins: SnowpackPlugin[];
    devOptions: {
        secure: boolean;
        hostname: string;
        port: number;
        out: string;
        fallback: string;
        open: string;
        bundle: boolean | undefined;
        hmr: boolean;
    };
    installOptions: {
        dest: string;
        env: EnvVarReplacements;
        treeshake?: boolean;
        installTypes: boolean;
        sourceMap?: boolean | 'inline';
        externalPackage: string[];
        namedExports: string[];
        rollup: {
            plugins: RollupPlugin[];
            dedupe?: string[];
        };
    };
    buildOptions: {
        baseUrl: string;
        webModulesUrl: string;
        clean: boolean;
        metaDir: string;
        minify: boolean;
    };
    _extensionMap: Record<string, string>;
    _bundler: SnowpackPlugin | undefined;
}
export interface CLIFlags extends Omit<Partial<SnowpackConfig['installOptions']>, 'env'> {
    help?: boolean;
    version?: boolean;
    reload?: boolean;
    config?: string;
    env?: string[];
    open?: string[];
    secure?: boolean;
}
export declare function createConfiguration(config: Partial<SnowpackConfig>): [ValidatorResult['errors'], undefined] | [null, SnowpackConfig];
export declare function loadAndValidateConfig(flags: CLIFlags, pkgManifest: any): SnowpackConfig;
export declare function removeLeadingSlash(path: string): string;
export declare function removeTrailingSlash(path: string): string;
export declare function addLeadingSlash(path: string): string;
export declare function addTrailingSlash(path: string): string;
